"""Window Parameters with instructions for generating windows."""
from pydantic import Field, constr
from typing import List

from honeybee_schema._base import NoExtraBaseModel


class ExtrudedBorder(NoExtraBaseModel):
    """Extruded borders over all windows in the wall."""

    type: constr(regex='^ExtrudedBorder$') = 'ExtrudedBorder'

    depth: float = Field(
        ...,
        gt=0,
        description='A number for the depth of the border.'
    )


class Overhang(NoExtraBaseModel):
    """A single overhang over an entire wall."""

    type: constr(regex='^Overhang$') = 'Overhang'

    depth: float = Field(
        ...,
        gt=0,
        description='A number for the overhang depth.'
    )

    angle: float = Field(
        0,
        ge=-90,
        le=90,
        description='A number between -90 and 90 for the for an angle to rotate the '
        'overhang in degrees. 0 indicates an overhang perpendicular to the wall. '
        'Positive values indicate a downward rotation. Negative values indicate '
        'an upward rotation.'
    )


class _LouversBase(NoExtraBaseModel):
    """Base class for for a series of louvered shades over a wall."""

    depth: float = Field(
        ...,
        gt=0,
        description='A number for the depth to extrude the louvers.'
    )

    offset: float = Field(
        0,
        ge=0,
        description='A number for the distance to louvers from the wall.'
    )

    angle: float = Field(
        0,
        ge=-90,
        le=90,
        description='A number between -90 and 90 for the for an angle to rotate the '
        'louvers in degrees. 0 indicates louvers perpendicular to the wall. '
        'Positive values indicate a downward rotation. Negative values indicate '
        'an upward rotation.'
    )

    contour_vector: List[float] = Field(
        [0, 1],
        min_items=2,
        max_items=2,
        description='A list of two float values representing the (x, y) of a 2D vector '
        'for the direction along which contours are generated. (0, 1) will generate '
        'horizontal contours, (1, 0) will generate vertical contours, and (1, 1) '
        'will generate diagonal contours.'
    )

    flip_start_side: bool = Field(
        False,
        description='Boolean to note whether the side the louvers start from should '
        'be flipped. Default is False to have contours on top or right. Setting '
        'to True will start contours on the bottom or left.'
    )


class LouversByDistance(_LouversBase):
    """A series of louvered Shades at a given distance between each louver."""

    type: constr(regex='^LouversByDistance$') = 'LouversByDistance'

    distance: float = Field(
        ...,
        gt=0,
        description='A number for the approximate distance between each louver.'
    )


class LouversByCount(_LouversBase):
    """A specific number of louvered Shades over a wall."""

    type: constr(regex='^LouversByCount$') = 'LouversByCount'

    louver_count: int = Field(
        ...,
        gt=0,
        description='A positive integer for the number of louvers to generate.'
    )



"""Model DOE-2 properties."""
from pydantic import Field, constr
from typing import Union

from honeybee_schema._base import NoExtraBaseModel
from honeybee_schema.altnumber import Autocalculate


class Room2DDoe2Properties(NoExtraBaseModel):

    type: constr(regex='^Room2DDoe2Properties$') = \
        'Room2DDoe2Properties'

    assigned_flow: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        description='A number for the design supply air flow rate for the zone '
        'the Room is assigned to (cfm). This establishes the minimum allowed '
        'design air flow. Note that the actual design flow may be larger. If '
        'Autocalculate, this parameter will not be written into the INP.'
    )

    flow_per_area: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        description='A number for the design supply air flow rate to the zone '
        'per unit floor area (cfm/ft2). If Autocalculate, this parameter will '
        'not be written into the INP.'
    )

    min_flow_ratio: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        le=1,
        description='A number between 0 and 1 for the minimum allowable zone '
        'air supply flow rate, expressed as a fraction of design flow rate. Applicable '
        'to variable-volume type systems only. If Autocalculate, this parameter will '
        'not be written into the INP.'
    )

    min_flow_per_area: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        description='A number for the minimum air flow per square foot of '
        'floor area (cfm/ft2). This is an alternative way of specifying the '
        'min_flow_ratio. If Autocalculate, this parameter will not be written '
        'into the INP.'
    )

    hmax_flow_ratio: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        le=1,
        description='A number between 0 and 1 for the ratio of the maximum (or fixed) '
        'heating airflow to the cooling airflow. The specific meaning varies according '
        'to the type of zone terminal. If Autocalculate, this parameter will '
        'not be written into the INP.'
    )


class ModelDoe2Properties(NoExtraBaseModel):

    type: constr(regex='^ModelDoe2Properties$') = \
        'ModelDoe2Properties'



"""Window Parameters with instructions for generating windows."""
from pydantic import Field, root_validator, constr, conlist, confloat
from typing import List

from honeybee_schema._base import NoExtraBaseModel


class _WindowParameterBase(NoExtraBaseModel):
    """Base class for all window parameters."""

    user_data: dict = Field(
        default=None,
        description='Optional dictionary of user data associated with the object.'
        'All keys and values of this dictionary should be of a standard data '
        'type to ensure correct serialization of the object (eg. str, float, '
        'int, list). When a list is used, each item in the list will be assigned '
        'to the generated Honeybee apertures.'
    )


class SingleWindow(_WindowParameterBase):
    """A single window in the wall center defined by a width * height."""

    type: constr(regex='^SingleWindow$') = 'SingleWindow'

    width: float = Field(
        ...,
        gt=0,
        description='A number for the window width. Note that, if this width is '
        'applied to a wall that is too narrow for this width, the generated '
        'window will automatically be shortened when it is applied to the wall. '
        'In this way, setting the width to be `float("inf")` will create '
        'parameters that always generate a ribbon window.'
    )

    height: float = Field(
        ...,
        gt=0,
        description='A number for the window height. Note that, if this height is '
        'applied to a wall that is too short for this height, the generated '
        'window will automatically be shortened when it is applied to the wall.'
    )

    sill_height: float = Field(
        1.0,
        gt=0,
        description='A number for the window sill height.'
    )


class SimpleWindowArea(_WindowParameterBase):
    """A single window defined by an absolute area."""

    type: constr(regex='^SimpleWindowArea$') = 'SimpleWindowArea'

    window_area: float = Field(
        ...,
        gt=0,
        description='A number for the window area in current model units. '
        'If this area is larger than the area of the Wall that it is applied'
        'to, the window will fill the parent Wall at a 99 percent ratio.'
    )

    rect_split: bool = Field(
        True,
        description='Boolean to note whether rectangular portions of base Face '
        'should be extracted before scaling them to create apertures. For '
        'pentagonal gabled geometries, this results in one rectangle and one '
        'triangle, which can often look more realistic and is a better input '
        'for engines like EnergyPlus that cannot model windows with more than 4 '
        'vertices. However, if a single pentagonal window is desired for such a '
        'gabled shape, this input can be set to False to produce such a result.'
    )


class SimpleWindowRatio(_WindowParameterBase):
    """A single window defined by an area ratio with the base surface."""

    type: constr(regex='^SimpleWindowRatio$') = 'SimpleWindowRatio'

    window_ratio: float = Field(
        ...,
        gt=0,
        lt=1,
        description='A number between 0 and 1 for the ratio between the window '
        'area and the parent wall surface area.'
    )

    rect_split: bool = Field(
        True,
        description='Boolean to note whether rectangular portions of base Face '
        'should be extracted before scaling them to create apertures. For '
        'pentagonal gabled geometries, this results in one rectangle and one '
        'triangle, which can often look more realistic and is a better input '
        'for engines like EnergyPlus that cannot model windows with more than 4 '
        'vertices. However, if a single pentagonal window is desired for such a '
        'gabled shape, this input can be set to False to produce such a result.'
    )


class RepeatingWindowRatio(_WindowParameterBase):
    """Repeating windows derived from an area ratio with the base wall."""

    type: constr(regex='^RepeatingWindowRatio$') = 'RepeatingWindowRatio'

    window_ratio: float = Field(
        ...,
        gt=0,
        lt=0.95,
        description='A number between 0 and 1 for the ratio between the window '
        'area and the parent wall surface area.'
    )

    window_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height of the windows. Note that, '
        'if the window ratio is too large for the height, the ratio will take '
        'precedence and the actual window_height will be larger than this value.'
    )

    sill_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height above the bottom edge of the '
        'wall to start the windows. Note that, if the ratio is too large '
        'for the height, the ratio will take precedence and the sill_height '
        'will be smaller than this value.'
    )

    horizontal_separation: float = Field(
        ...,
        ge=0,
        description='A number for the target separation between individual window '
        'centerlines.  If this number is larger than the parent rectangle base, '
        'only one window will be produced.'
    )

    vertical_separation: float = Field(
        0,
        ge=0,
        description='An optional number to create a single vertical separation '
        'between top and bottom windows.'
    )


class RepeatingWindowWidthHeight(_WindowParameterBase):
    """Repeating rectangular windows of a fixed width and height."""

    type: constr(regex='^RepeatingWindowWidthHeight$') = 'RepeatingWindowWidthHeight'

    window_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height of the windows. Note that, '
        'if the window_height is larger than the height of the wall, '
        'the generated windows will have a height equal to the wall height '
        'in order to avoid having windows extend outside the wall face.'
    )

    window_width: float = Field(
        ...,
        gt=0,
        description='A number for the target width of the windows. '
        'Note that, if the window_width is larger than the width of the wall, '
        'the generated windows will have a width equal to the wall width '
        'in order to avoid having windows extend outside the wall face.'
    )

    sill_height: float = Field(
        ...,
        gt=0,
        description='A number for the target height above the bottom edge of the '
        'wall to start the windows. If the window_height is too large for the '
        'sill_height to fit within the rectangle, the window_height will take '
        'precedence.'
    )

    horizontal_separation: float = Field(
        ...,
        ge=0,
        description='A number for the target separation between individual window '
        'centerlines.  If this number is larger than the parent rectangle base, '
        'only one window will be produced.'
    )


class RectangularWindows(_WindowParameterBase):
    """Several rectangular windows, defined by origin, width and height."""

    type: constr(regex='^RectangularWindows$') = 'RectangularWindows'

    origins: List[conlist(confloat(gt=0), min_items=2, max_items=2)] = Field(
        ...,
        min_items=1,
        description='An array of 2D points within the plane of the wall for the origin '
        'of each window. Each point should be a list of 2 (x, y) values. The '
        'wall plane is assumed to have an origin at the first point of the wall '
        'segment and an X-axis extending along the length of the segment. The '
        'wall plane Y-axis always points upwards. Therefore, both X and Y '
        'values of each origin point should be positive.'
    )

    widths: List[confloat(gt=0)] = Field(
        ...,
        min_items=1,
        description='An array of positive numbers for the window widths. '
        'The length of this list must match the length of the origins.'
    )

    heights: List[confloat(gt=0)] = Field(
        ...,
        min_items=1,
        description='An array of positive numbers for the window heights. '
        'The length of this list must match the length of the origins.'
    )

    are_doors: List[bool] = Field(
        default=None,
        description='An array of booleans that align with the origins and note '
        'whether each of the geometries represents a door (True) or a window (False). '
        'If None, it will be assumed that all geometries represent windows and '
        'they will be translated to Apertures in any resulting Honeybee model.'
    )

    @root_validator
    def check_aligned_lists(cls, values):
        "Ensure length of origins, widths and heights match."
        origins = values.get('origins')
        widths = values.get('widths')
        heights = values.get('heights')

        assert len(origins) == len(widths) == len(heights), 'Length of ' \
            'RectangularWindows origins, widths, and heights ' \
            'must match. origins: {}, widths: {}, heights: {}'.format(
                len(origins), len(widths), len(heights))

        return values


class DetailedWindows(_WindowParameterBase):
    """Several detailed windows defined by 2D Polygons (lists of 2D vertices)."""

    type: constr(regex='^DetailedWindows$') = 'DetailedWindows'

    polygons: List[
        conlist(conlist(confloat(gt=0), min_items=2, max_items=3), min_items=3)
    ] = Field(
        ...,
        description='An array of arrays with each sub-array representing a polygonal '
        'boundary of a window. Each sub-array should consist of arrays '
        'representing points, which can either contain 2 values (indicating they '
        'are 2D vertices within the plane of a parent wall segment) or they can '
        'contain 3 values (indicating they are 3D world coordinates). '
        'For 2D points, the wall plane is assumed to have an origin at the '
        'first point of the wall segment and an X-axis extending along the '
        'length of the segment. The wall plane Y-axis always points upwards. '
        'Therefore, both X and Y values of each point in the polygon should '
        'always be positive. Some sample code to convert from 2D vertices to '
        '2D vertices in the plane of the wall can be found here: '
        'https://www.ladybug.tools/dragonfly-core/docs/dragonfly.windowparameter.'
        'html#dragonfly.windowparameter.DetailedWindows'
    )

    are_doors: List[bool] = Field(
        default=None,
        description='An array of booleans that align with the polygons and note '
        'whether each of the polygons represents a door (True) or a window (False). '
        'If None, it will be assumed that all polygons represent windows and '
        'they will be translated to Apertures in any resulting Honeybee model.'
    )



"""Model energy properties."""
from pydantic import Field, constr
from typing import List, Union

from honeybee_schema._base import NoExtraBaseModel
from honeybee_schema.energy.constructionset import ConstructionSetAbridged, \
    ConstructionSet
from honeybee_schema.energy.programtype import ProgramTypeAbridged, ProgramType
from honeybee_schema.energy.hvac.idealair import IdealAirSystemAbridged
from honeybee_schema.energy.hvac.allair import VAV, PVAV, PSZ, PTAC, ForcedAirFurnace
from honeybee_schema.energy.hvac.doas import FCUwithDOASAbridged, \
    WSHPwithDOASAbridged, VRFwithDOASAbridged, RadiantwithDOASAbridged
from honeybee_schema.energy.hvac.heatcool import FCU, WSHP, VRF, Baseboard, \
    EvaporativeCooler, Residential, WindowAC, GasUnitHeater, Radiant
from honeybee_schema.energy.hvac.detailed import DetailedHVAC
from honeybee_schema.energy.ventcool import VentilationControlAbridged, \
    VentilationOpening
from honeybee_schema.energy.load import ProcessAbridged
from honeybee_schema.energy.shw import SHWSystem
from honeybee_schema.energy.global_constructionset import GlobalConstructionSet

from honeybee_schema.energy.construction import OpaqueConstructionAbridged, \
    WindowConstructionAbridged, ShadeConstruction, AirBoundaryConstructionAbridged, \
    OpaqueConstruction, WindowConstruction, AirBoundaryConstruction
from honeybee_schema.energy.material import EnergyMaterial, EnergyMaterialNoMass, \
    EnergyMaterialVegetation, EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom, \
    EnergyWindowMaterialGasMixture, EnergyWindowFrame, \
    EnergyWindowMaterialSimpleGlazSys, EnergyWindowMaterialGlazing, \
    EnergyWindowMaterialShade, EnergyWindowMaterialBlind
from honeybee_schema.energy.schedule import ScheduleTypeLimit, ScheduleRulesetAbridged, \
    ScheduleFixedIntervalAbridged, ScheduleRuleset, ScheduleFixedInterval


class Room2DEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^Room2DEnergyPropertiesAbridged$') = \
        'Room2DEnergyPropertiesAbridged'

    construction_set: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Name of a ConstructionSet to specify all constructions for '
        'the Room2D. If None, the Room2D will use the Story or Building '
        'construction_set or the Model global_construction_set. Any ConstructionSet '
        'assigned here will override those assigned to these objects.'
    )

    program_type: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Name of a ProgramType to specify all schedules and loads '
        'for the Room2D. If None, the Room2D will have no loads or setpoints.'
    )

    hvac: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An optional identifier of a HVAC system (such as an IdealAirSystem)'
        ' that specifies how the Room2D is conditioned. If None, it will be assumed '
        'that the Room2D is not conditioned.'
    )

    shw: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An optional identifier of a Service Hot Water (SHW) system '
        'that specifies how the hot water load of the Room is met. If None, the hot '
        'water load will be met with a generic system that only measures thermal load'
        'and does not account for system efficiencies.'
    )

    window_vent_control: VentilationControlAbridged = Field(
        default=None,
        description='An optional VentilationControl object to dictate the opening '
        'of windows. If None, the windows will never open.'
    )

    window_vent_opening: VentilationOpening = Field(
        default=None,
        description='An optional VentilationOpening to specify the operable '
        'portion of all windows of the Room2D. If None, the windows will never open.'
    )

    process_loads: List[ProcessAbridged] = Field(
        default=None,
        description='An optional list of Process objects for process loads within '
        'the room. These can represent wood burning fireplaces, kilns, manufacturing '
        'equipment, and various industrial processes. They can also be used to '
        'represent certain pieces of equipment to be separated from the other '
        'end uses, such as MRI machines, theatrical lighting, and elevators.'
    )


class StoryEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^StoryEnergyPropertiesAbridged$') = \
        'StoryEnergyPropertiesAbridged'

    construction_set: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Name of a ConstructionSet to specify all constructions for '
        'the Story. If None, the Story will use the Building construction_set '
        'or the Model global_construction_set. Any ConstructionSet '
        'assigned here will override those assigned to these objects.'
    )


class BuildingEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^BuildingEnergyPropertiesAbridged$') = \
        'BuildingEnergyPropertiesAbridged'

    construction_set: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Name of a ConstructionSet to specify all constructions for '
        'the Building. If None, the Model global_construction_set will be used.'
    )

    ceiling_plenum_construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of an OpaqueConstruction for the bottoms of '
        'ceiling plenums. Materials should be ordered from the plenum side '
        'to the room side. By default, this is a simple acoustic tile construction.'
    )

    floor_plenum_construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of an OpaqueConstruction for the tops of floor '
        'plenums. Materials should be ordered from the plenum side '
        'to the room side. By default, this is a simple wood plank construction.'
    )


class ContextShadeEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^ContextShadeEnergyPropertiesAbridged$') = \
        'ContextShadeEnergyPropertiesAbridged'

    construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Name of a ShadeConstruction to set the reflectance and '
        'specularity of the ContextShade. If None, the the EnergyPlus default '
        'of 0.2 diffuse reflectance will be used.'
    )

    transmittance_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Name of a schedule to set the transmittance of the ContextShade, '
        'which can vary throughout the simulation. If None, the ContextShade will '
        'be completely opaque.'
    )


class ModelEnergyProperties(NoExtraBaseModel):

    type: constr(regex='^ModelEnergyProperties$') = 'ModelEnergyProperties'

    global_construction_set: GlobalConstructionSet = Field(
        default=GlobalConstructionSet(),
        description='Global Energy construction set.',
        readOnly=True
    )

    construction_sets: List[Union[ConstructionSetAbridged, ConstructionSet]] = Field(
        default=None,
        description='List of all ConstructionSets in the Model.'
    )

    constructions: List[
        Union[
            OpaqueConstructionAbridged, WindowConstructionAbridged,
            ShadeConstruction, AirBoundaryConstructionAbridged,
            OpaqueConstruction, WindowConstruction, AirBoundaryConstruction
        ]
    ] = Field(
        default=None,
        description='A list of all unique constructions in the model. This includes '
        'constructions across all the Model construction_sets.'
    )

    materials: List[
        Union[
            EnergyMaterial, EnergyMaterialNoMass, EnergyMaterialVegetation,
            EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom,
            EnergyWindowMaterialGasMixture, EnergyWindowFrame,
            EnergyWindowMaterialSimpleGlazSys, EnergyWindowMaterialGlazing,
            EnergyWindowMaterialBlind, EnergyWindowMaterialShade
        ]
    ] = Field(
        default=None,
        description='A list of all unique materials in the model. This includes '
        'materials needed to make the Model constructions.'
    )

    hvacs: List[
        Union[
            IdealAirSystemAbridged, VAV, PVAV, PSZ, PTAC, ForcedAirFurnace,
            FCUwithDOASAbridged, WSHPwithDOASAbridged, VRFwithDOASAbridged,
            RadiantwithDOASAbridged, FCU, WSHP, VRF, Baseboard, EvaporativeCooler,
            Residential, WindowAC, GasUnitHeater, Radiant, DetailedHVAC
        ]
    ] = Field(
        default=None,
        description='List of all HVAC systems in the Model.'
    )

    shws: List[SHWSystem] = Field(
        default=None,
        description='List of all Service Hot Water (SHW) systems in the Model.'
    )

    program_types: List[Union[ProgramTypeAbridged, ProgramType]] = Field(
        default=None,
        description='List of all ProgramTypes in the Model.'
    )

    schedules: List[Union[ScheduleRulesetAbridged, ScheduleFixedIntervalAbridged,
                          ScheduleRuleset, ScheduleFixedInterval]] = Field(
        default=None,
        description='A list of all unique schedules in the model. This includes '
        'schedules across all HVAC systems, ProgramTypes and ContextShades.'
    )

    schedule_type_limits: List[ScheduleTypeLimit] = Field(
        default=None,
        description='A list of all unique ScheduleTypeLimits in the model. This '
        'all ScheduleTypeLimits needed to make the Model schedules.'
    )



"""Model schema and the 3 geometry objects that define it."""
from pydantic import BaseModel, Field, root_validator, constr, conlist
from typing import List, Union
from enum import Enum

from honeybee_schema._base import IDdBaseModel
from honeybee_schema.model import Room, Face3D, Mesh3D, Units
from honeybee_schema.boundarycondition import Ground, Outdoors, Surface, \
    Adiabatic, OtherSideTemperature
from honeybee_schema.altnumber import Autocalculate

from .window_parameter import SingleWindow, SimpleWindowArea, SimpleWindowRatio, \
    RepeatingWindowRatio, RectangularWindows, DetailedWindows
from .shading_parameter import ExtrudedBorder, Overhang, LouversByDistance, \
    LouversByCount
from .skylight_parameter import GriddedSkylightArea, GriddedSkylightRatio, \
    DetailedSkylights
from .roof import RoofSpecification
from .energy.properties import Room2DEnergyPropertiesAbridged, \
    StoryEnergyPropertiesAbridged, BuildingEnergyPropertiesAbridged, \
    ContextShadeEnergyPropertiesAbridged, ModelEnergyProperties
from .radiance.properties import Room2DRadiancePropertiesAbridged, \
    StoryRadiancePropertiesAbridged, BuildingRadiancePropertiesAbridged, \
    ContextShadeRadiancePropertiesAbridged, ModelRadianceProperties
from .doe2.properties import Room2DDoe2Properties, ModelDoe2Properties
from .comparison.properties import Room2DComparisonProperties, ModelComparisonProperties


class Room2DPropertiesAbridged(BaseModel):

    type: constr(regex='^Room2DPropertiesAbridged$') = 'Room2DPropertiesAbridged'

    energy: Room2DEnergyPropertiesAbridged = Field(
        default=None
    )

    radiance: Room2DRadiancePropertiesAbridged = Field(
        default=None
    )

    doe2: Room2DDoe2Properties = Field(
        default=None
    )

    comparison: Room2DComparisonProperties = Field(
        default=None
    )


class Room2D(IDdBaseModel):

    type: constr(regex='^Room2D$') = 'Room2D'

    floor_boundary: List[conlist(float, min_items=2, max_items=2)] = Field(
        ...,
        min_items=3,
        description='A list of 2D points representing the outer boundary vertices of '
        'the Room2D. The list should include at least 3 points and each point '
        'should be a list of 2 (x, y) values.'
    )

    floor_holes: List[conlist(conlist(float, min_items=2, max_items=2), min_items=3)] \
        = Field(
        None,
        description='Optional list of lists with one list for each hole in the floor '
        'plate. Each hole should be a list of at least 2 points and each point a list '
        'of 2 (x, y) values. If None, it will be assumed that there are no '
        'holes in the floor plate.'
    )

    floor_height: float = Field(
        ...,
        description='A number to indicate the height of the floor plane in the Z axis.'
    )

    floor_to_ceiling_height: float = Field(
        ...,
        description='A number for the distance between the floor and the ceiling.'
    )

    is_ground_contact: bool = Field(
        False,
        description='A boolean noting whether this Room2D has its floor in contact '
        'with the ground.'
    )

    is_top_exposed: bool = Field(
        False,
        description='A boolean noting whether this Room2D has its ceiling exposed '
        'to the outdoors.'
    )

    has_floor: bool = Field(
        True,
        description='A boolean for whether the room has a Floor (True) or an '
        'AirBoundary (False). If False, this property will only be meaningful if the '
        'model is translated to Honeybee with ceiling adjacency solved and there '
        'is a Room2D below this one with a has_ceiling property set to False.'
    )

    has_ceiling: bool = Field(
        True,
        description='A boolean for whether the room has a RoofCeiling (True) or an '
        'AirBoundary (False). If False, this property will only be meaningful if the '
        'model is translated to Honeybee with ceiling adjacency solved and there '
        'is a Room2D above this one with a has_floor property set to False.'
    )

    ceiling_plenum_depth: float = Field(
        0,
        ge=0,
        description='A number for the depth that a ceiling plenum extends into '
        'the room. Setting this to a positive value will result in a separate '
        'plenum room being split off of the Room2D volume during translation '
        'from Dragonfly to Honeybee. The bottom of this ceiling plenum will '
        'always be at this Room2D ceiling height minus the value specified here. '
        'Setting this to zero indicates that the room has no ceiling plenum.'
    )

    floor_plenum_depth: float = Field(
        0,
        ge=0,
        description='A number for the depth that a floor plenum extends into '
        'the room. Setting this to a positive value will result in a separate '
        'plenum room being split off of the Room2D volume during translation '
        'from Dragonfly to Honeybee. The top of this floor plenum will always '
        'be at this Room2D floor height plus the value specified here. '
        'Setting this to zero indicates that the room has no floor plenum.'
    )

    zone: str = Field(
        default=None,
        description='Text string for for the zone identifier to which this Room2D '
        ' belongs. Room2Ds sharing the same zone identifier are considered part of the '
        'same zone in a Building. If the zone identifier has not been specified, it will '
        'be the same as the Room2D identifier in the destination engine. Note that this '
        'property has no character restrictions.'
    )

    boundary_conditions: List[
        Union[Ground, Outdoors, Surface, Adiabatic, OtherSideTemperature]
    ] = Field(
        default=None,
        description='A list of boundary conditions that match the number of segments '
        'in the input floor_geometry + floor_holes. These will be used to assign '
        'boundary conditions to each of the walls of the Room in the resulting '
        'model. Their order should align with the order of segments in the '
        'floor_boundary and then with each hole segment. If None, all boundary '
        'conditions will be Outdoors or Ground depending on whether ceiling '
        'height of the room is at or below 0 (the assumed ground plane).'
    )

    window_parameters: List[Union[
        None, SingleWindow, SimpleWindowArea, SimpleWindowRatio, RepeatingWindowRatio,
        RectangularWindows, DetailedWindows
    ]] = Field(
        default=None,
        description='A list of WindowParameter objects that dictate how the window '
        'geometries will be generated for each of the walls. If None, no windows '
        'will exist over the entire Room2D.'
    )

    shading_parameters: List[Union[
        None, ExtrudedBorder, Overhang, LouversByDistance, LouversByCount
    ]] = Field(
        default=None,
        description='A list of ShadingParameter objects that dictate how the shade '
        'geometries will be generated for each of the walls. If None, no shades '
        'will exist over the entire Room2D.'
    )

    air_boundaries: List[bool] = Field(
        default=None,
        description='A list of booleans for whether each wall has an air boundary type. '
        'False values indicate a standard opaque type while True values indicate '
        'an AirBoundary type. All walls will be False by default. Note that any '
        'walls with a True air boundary must have a Surface boundary condition '
        'without any windows.'
    )

    skylight_parameters: Union[
        None, GriddedSkylightArea, GriddedSkylightRatio, DetailedSkylights
    ] = Field(
        default=None,
        description='A SkylightParameter object describing how to generate skylights. '
        'If None, no skylights will exist on the Room2D.'
    )

    properties: Room2DPropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )

    @root_validator
    def check_segment_count(cls, values):
        "Ensure len of boundary_conditions, window par, shading par match segment count."
        floor_bound = values.get('floor_boundary')
        floor_holes = values.get('floor_holes')
        bcs = values.get('boundary_conditions')
        win_par = values.get('window_parameters')
        shd_par = values.get('shading_parameters')
        air_bnd = values.get('air_boundaries')

        seg_count = len(floor_bound) if floor_holes is None\
            else len(floor_bound) + sum(len(hole) for hole in floor_holes)

        if bcs is not None:
            assert len(bcs) == seg_count, 'Length of Room2D boundary_conditions ' \
                'must match number of floor segments. {} != {}'.format(
                    len(bcs), seg_count)
        if win_par is not None:
            assert len(win_par) == seg_count, 'Length of Room2D window_parameters ' \
                'must match number of floor segments. {} != {}'.format(
                    len(win_par), seg_count)
        if shd_par is not None:
            assert len(shd_par) == seg_count, 'Length of Room2D shading_parameters ' \
                'must match number of floor segments. {} != {}'.format(
                    len(shd_par), seg_count)
        if air_bnd is not None:
            assert len(air_bnd) == seg_count, 'Length of Room2D air_boundaries ' \
                'must match number of floor segments. {} != {}'.format(
                    len(air_bnd), seg_count)
        return values


class StoryType(str, Enum):
    standard = 'Standard'
    ceiling_plenum = 'CeilingPlenum'
    floor_plenum = 'FloorPlenum'


class StoryPropertiesAbridged(BaseModel):

    type: constr(regex='^StoryPropertiesAbridged$') = 'StoryPropertiesAbridged'

    energy: StoryEnergyPropertiesAbridged = Field(
        default=None
    )

    radiance: StoryRadiancePropertiesAbridged = Field(
        default=None
    )


class Story(IDdBaseModel):

    type: constr(regex='^Story$') = 'Story'

    room_2ds: List[Room2D] = Field(
        ...,
        description='An array of dragonfly Room2D objects that together form an '
        'entire story of a building.'
    )

    floor_to_floor_height: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        description='A number for the distance from the floor plate of this story '
        'to the floor of the story above this one (if it exists). If Autocalculate, '
        'this value will be the maximum floor_to_ceiling_height of the input room_2ds.'
    )

    floor_height: Union[Autocalculate, float] = Field(
        Autocalculate(),
        description='A number to indicate the height of the floor plane in the Z axis.'
        'If Autocalculate, this will be the minimum floor height of all the room_2ds, '
        'which is suitable for cases where there are no floor plenums.'
    )

    multiplier: int = Field(
        1,
        ge=1,
        description='An integer that denotes the number of times that this Story is '
        'repeated over the height of the building.'
    )

    roof: RoofSpecification = Field(
        default=None,
        description='An optional RoofSpecification object containing geometry '
        'for generating sloped roofs over the Story. The RoofSpecification will only '
        'affect the child Room2Ds that have a True is_top_exposed property '
        'and it will only be utilized in translation to Honeybee when the Story '
        'multiplier is 1. If None, all Room2D ceilings will be flat.'
    )

    story_type: StoryType = Field(
        StoryType.standard,
        description='Text to indicate the type of story. Stories that are plenums '
        'are translated to Honeybee with excluded floor areas.'
    )

    properties: StoryPropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )


class BuildingPropertiesAbridged(BaseModel):

    type: constr(regex='^BuildingPropertiesAbridged$') = 'BuildingPropertiesAbridged'

    energy: BuildingEnergyPropertiesAbridged = Field(
        default=None
    )

    radiance: BuildingRadiancePropertiesAbridged = Field(
        default=None
    )


class Building(IDdBaseModel):

    type: constr(regex='^Building$') = 'Building'

    unique_stories: List[Story] = Field(
        default=None,
        description='An array of unique dragonfly Story objects that together form '
        'the entire building. Stories should generally be ordered from lowest '
        'floor to highest floor, though this is not required. Note that, if a '
        'given Story is repeated several times over the height of the building '
        'and this is represented by the multiplier, the unique story included in this '
        'list should be the first (lowest) story of the repeated floors.'
    )

    room_3ds: List[Room] = Field(
        default=None,
        description='An optional array of 3D Honeybee Room objects for additional '
        'Rooms that are a part of the Building but are not represented within '
        'the unique_stories. This is useful when there are parts of the Building '
        'geometry that cannot easily be represented with the extruded floor '
        'plate and sloped roof assumptions that underlie Dragonfly Room2Ds '
        'and RoofSpecification. Cases where this input is most useful include '
        'sloped walls and certain types of domed roofs that become tedious to '
        'implement with RoofSpecification. Matching the Honeybee Room.story '
        'property to the Dragonfly Story.display_name of an object within the '
        'unique_stories will effectively place the Honeybee Room on that Story '
        'for the purposes of floor_area, exterior_wall_area, etc. However, note '
        'that the Honeybee Room.multiplier property takes precedence over '
        'whatever multiplier is assigned to the Dragonfly Story that the '
        'Room.story may reference. (Default: None).'
    )

    roof: RoofSpecification = Field(
        default=None,
        description='An optional RoofSpecification object that provides an '
        'alternative way to describe roof geometry over rooms (instead of '
        'specifying roofs on each story). If trying to decide between the two, '
        'specifying geometry on each story is closer to how dragonfly natively '
        'works with roof geometry as it relates to rooms. However, when it is '
        'unknown which roof geometries correspond to which stories, this Building-level '
        'attribute may be used and each roof geometry will automatically be added '
        'to the best Story in the Building upon serialization to Python. This '
        'automatic assignment will happen by checking for overlaps between the '
        'Story Room2Ds and the Roof geometry in plan. When a given roof geometry '
        'overlaps with several Stories, the top-most Story will get the roof geometry '
        'assigned to it unless this top Story has a floor_height above the roof '
        'geometry, in which case the next highest story will be checked until a '
        'compatible one is found.'
    )

    properties: BuildingPropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )


class ContextShadePropertiesAbridged(BaseModel):

    type: constr(regex='^ContextShadePropertiesAbridged$') = \
        'ContextShadePropertiesAbridged'

    energy: ContextShadeEnergyPropertiesAbridged = Field(
        default=None
    )

    radiance: ContextShadeRadiancePropertiesAbridged = Field(
        default=None
    )


class ContextShade(IDdBaseModel):

    type: constr(regex='^ContextShade$') = 'ContextShade'

    geometry: List[Union[Face3D, Mesh3D]] = Field(
        ...,
        description='An array of planar Face3Ds and or Mesh3Ds that together '
        'represent the context shade.'
    )

    is_detached: bool = Field(
        True,
        description='Boolean to note whether this shade is detached from any of '
        'the other geometry in the model. Cases where this should be True include '
        'shade representing surrounding buildings or context.'
    )

    properties: ContextShadePropertiesAbridged = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )


class ModelProperties(BaseModel):

    type: constr(regex='^ModelProperties$') = 'ModelProperties'

    energy: ModelEnergyProperties = Field(
        default=None
    )

    radiance: ModelRadianceProperties = Field(
        default=None
    )

    doe2: ModelDoe2Properties = Field(
        default=None
    )

    comparison: ModelComparisonProperties = Field(
        default=None
    )


class Model(IDdBaseModel):

    type: constr(regex='^Model$') = 'Model'

    version: str = Field(
        default='0.0.0',
        regex=r'([0-9]+)\.([0-9]+)\.([0-9]+)',
        description='Text string for the current version of the schema.'
    )

    buildings: List[Building] = Field(
        None,
        description='A list of Buildings in the model.'
    )

    context_shades: List[ContextShade] = Field(
        None,
        description='A list of ContextShades in the model.'
    )

    units: Units = Field(
        default=Units.meters,
        description='Text indicating the units in which the model geometry exists. '
        'This is used to scale the geometry to the correct units for simulation '
        'engines like EnergyPlus, which requires all geometry be in meters.'
    )

    tolerance: float = Field(
        default=0.01,
        ge=0,
        description='The maximum difference between x, y, and z values at which '
        'vertices are considered equivalent. This value should be in the Model '
        'units and is used in a variety of checks and operations. A value of 0 '
        'will result in bypassing all checks so it is recommended that this always '
        'be a positive number when checks have not already been performed '
        'on a Model. The default of 0.01 is suitable for models in meters.'
    )

    angle_tolerance: float = Field(
        default=1.0,
        ge=0,
        description='The max angle difference in degrees that vertices are '
        'allowed to differ from one another in order to consider them colinear. '
        'This value is used in a variety of checks and operations that can be '
        'performed on geometry. A value of 0 will result in no checks and '
        'an inability to perform certain operations so it is recommended that '
        'this always be a positive number when checks have not already '
        'been performed on a given Model.'
    )

    reference_vector: List[float] = Field(
        None,
        description='A n optional list of 3 (x, y, z) values that describe a Vector3D '
        'relating the model to an original source coordinate system. Setting a value '
        'here is useful if the model has been moved from its original location '
        'and there may be future operations of merging geometry from the original '
        'source system.',
        min_items=3,
        max_items=3
    )

    properties: ModelProperties = Field(
        ...,
        description='Extension properties for particular simulation engines '
        '(Radiance, EnergyPlus).'
    )
